#include <cstring>
#include <cstdio>

#include "networkClient.hpp"

template <>
EGCiphertext get_committed_val_from_file<EGCiphertext>(
    struct synchronization_tool *sync,
    char *filename,
    Proof& pi)
{
    std::unique_lock<std::mutex> lck(sync->mtx);
    std::ifstream valFile(filename);

    EGCiphertext retval;
    valFile >> pi;
    valFile >> retval;

    return retval;
}

template <>
std::vector<TwistBipoint> get_committed_val_from_file<std::vector<TwistBipoint>>(
    struct synchronization_tool *sync,
    char *filename,
    Proof& pi)
{
    std::unique_lock<std::mutex> lck(sync->mtx);
    std::ifstream valFile(filename);

    valFile >> pi;

    std::vector<TwistBipoint> retval;

    BinarySizeT sizeOfVector;
    for (size_t i = 0; i < sizeOfVector.val(); i++)
    {
        TwistBipoint currVote;
        valFile >> currVote;

        retval.push_back(currVote);
    }

    return retval;
}

template <typename T>
T get_first_committed_val(
    const std::string& server,
    Proof& pi,
    const Twistpoint& shortTermPublicKey,
    const char *firstUri)
{
    struct synchronization_tool sync;

    std::stringstream buffer;
    std::string data;

    buffer << shortTermPublicKey;
    data = buffer.str();

    bool flag = false;
    while (!flag)
    {
        struct mg_connection *conn =
            mg_connect_websocket_client(
                server.c_str(),
                PRSONA_PORT,
                USE_SSL,
                NULL,
                0,
                firstUri,
                "null",
                file_websocket_data_handler,
                file_websocket_close_handler,
                (void *) &sync);

        if (!conn)
        {
            std::cerr << "Trouble getting encrypted score from server at " << server << std::endl;
            continue;
        }

        std::unique_lock<std::mutex> lck(sync.mtx);
        char *filename = set_temp_filename(conn);
        sync.val = 0;

        mg_websocket_client_write(
            conn,
            MG_WEBSOCKET_OPCODE_BINARY,
            data.c_str(),
            data.length());

        mg_websocket_client_write(
            conn,
            MG_WEBSOCKET_OPCODE_DATACOMPLETE,
            "",
            0);

        while (!sync.val)
            sync.cv.wait(lck)

        mg_close_connection(conn);

        flag = true;
    }

    T retval = get_committed_val_from_file<T>(filename, pi);

    remove(filename);
    delete filename;

    return retval;
}

Proof get_commitment_from_file(
    struct synchronization_tool *sync,
    char *filename)
{
    std::unique_lock<std::mutex> lck(sync->mtx);
    std::ifstream scoreFile(filename);

    Proof retval;
    scoreFile >> retval;

    return retval;
}

void get_additional_commitment(
    const std::vector<std::string>& servers,
    const std::string& skip,
    std::vector<Proof>& pi,
    const Twistpoint& shortTermPublicKey,
    char *commitUri)
{
    std::vector<char *> commitmentFilenames;
    std::vector<struct synchronization_tool *> commitmentSyncs;

    for (size_t i = 0; i < servers.size(); i++)
    {
        if (servers[i] == skip)
            continue;

        struct synchronization_tool *currSync = new struct synchronization_tool;
        commitmentSyncs.push_back(currSync);

        bool flag = false;
        while (!flag)
        {
            struct mg_connection *conn =
                mg_connect_websocket_client(
                    servers[i].c_str(),
                    PRSONA_PORT,
                    USE_SSL,
                    NULL,
                    0,
                    commitUri,
                    "null",
                    file_websocket_data_handler,
                    file_websocket_close_handler,
                    (void *) currSync);

            if (!conn)
            {
                std::cerr << "Trouble getting commitment from server at " << servers[i] << std::endl;
                continue;
            }

            std::unique_lock<std::mutex> lck(currSync->mtx);
            currSync->val = 0;
            commitmentFilenames.push_back(set_temp_filename(conn));

            mg_websocket_client_write(
                conn,
                MG_WEBSOCKET_OPCODE_DATACOMPLETE,
                "",
                0);

            while (!currSync->val)
                currSync->cv.wait(lck)

            mg_close_connection(conn);

            flag = true;
        }  
    }
    
    for (size_t i = 0; i < commitmentFilenames.size(); i++)
    {
        pi.push_back(
            get_commitment_from_file(
                commitmentSyncs[i], commitmentFilenames[i]));

        delete commitmentSyncs[i];

        remove(commitmentFilenames[i]);
        delete commitmentFilenames[i];
    }
}

template <typename T>
T get_server_committed_val(
    std::default_random_engine *generator,
    const std::vector<std::string>& serverIPs,
    std::vector<Proof>& pi,
    const Twistpoint& shortTermPublicKey,
    const char *firstUri
    const char *commitUri)
{
    std::uniform_int_distribution<size_t> distribution(0, serverIPs.size() - 1);
    size_t whichServer = distribution(*generator);

    pi.clear();

    Proof firstProof;
    T retval =
        get_first_committed_val<T>(
            serverIPs[whichServer],
            firstProof,
            shortTermPublicKey
            firstUri);

    pi.push_back(firstProof);

    get_additional_commitment(
        serverIPs,
        serverIPs[whichServer],
        pi,
        shortTermPublicKey,
        commitUri);

    return retval;
}

Twistpoint get_generator_from_file(
    struct synchronization_tool *sync,
    const char *filename,
    std::vector<Proof>& pi)
{
    std::ifstream genFile(filename);

    BinarySizeT sizeOfVector;
    genFile >> sizeOfVector;

    for (size_t i = 0; i < sizeOfVector.val(); i++)
    {
        Proof currProof;
        genFile >> currProof;

        pi.push_back(currProof);
    }

    Twistpoint retval;
    genFile >> retval;

    return retval;
}

Twistpoint get_generator(
    std::default_random_engine *randomGenerator,
    const std::vector<std::string>& serverIPs,
    std::vector<Proof>& pi,
    bool fresh)
{
    struct synchronization_tool sync;

    std::uniform_int_distribution<size_t> distribution(0, serverIPs.size() - 1);
    size_t whichServer = distribution(*randomGenerator);

    pi.clear();

    const char* whichUri = (fresh ? FRESH_GEN_URI : BLIND_GEN_URI);

    bool flag = false;
    while (!flag)
    {
        struct mg_connection *conn = mg_connect_websocket_client(
            serverIPs[whichServer].c_str(),
            PRSONA_PORT,
            USE_SSL,
            NULL,
            0,
            whichUri,
            "null",
            file_websocket_data_handler,
            file_websocket_close_handler,
            (void *) &sync);

        if (!conn)
        {
            cerr << "Couldn't obtain BGN details" << endl;
            return 1;
        }

        unique_lock<mutex> lck(sync.mtx);
        char *filename = set_temp_filename(conn);
        sync.val = 0;

        mg_websocket_client_write(
            conn,
            MG_WEBSOCKET_OPCODE_DATACOMPLETE,
            "",
            0);

        while (!sync.val)
            sync.cv.wait(lck);

        mg_close_connection(conn);

        flag = true;
    }

    Twistpoint retval =
        get_generator_from_file(&sync, filename, pi);

    remove(filename);
    delete filename;

    return retval;
}

PrsonaClientWebSocketHandler::PrsonaClientWebSocketHandler(
    const PrsonaClient *prsonaClient, 
    const std::vector<std::string>& serverIPs,
    const std::string& selfIP,
    const std::default_random_engine *generator)
: prsonaClient(prsonaClient), serverIPs(serverIPs),
    selfIP(selfIP), generator(generator)
{ /* */ }

virtual bool PrsonaClientWebSocketHandler::handleConnection(
    CivetServer *server,
    const struct mg_connection *conn)
{
    const struct mg_request_info *info = mg_get_request_info(conn);
    
    flag = (info->query_string && info->query_string[0] == PRSONA_VERIFY_REPUTATION_PROOF);

    return flag;
}

virtual void PrsonaClientWebSocketHandler::handleReadyState(
    CivetServer *server,
    struct mg_connection *conn)
{
    const struct mg_request_info *info = mg_get_request_info(conn);

    switch (info->query_string[0])
    {
        case PRSONA_VERIFY_REPUTATION_PROOF:
            set_temp_filename(conn);
            break;

        default:
            mg_set_user_connection_data(conn, NULL);
            break;
    }
}

virtual bool PrsonaClientWebSocketHandler::handleData(
    CivetServer *server,
    struct mg_connection *conn,
    int bits,
    char *data,
    size_t data_len)
{
    char *filename = (char *) mg_get_user_connection_data(conn);

    if ((bits & 0xf) == MG_WEBSOCKET_OPCODE_DATACOMPLETE)
    {
        generate_response(conn, filename);
        return false;
    }

    if ((bits & 0xf) != MG_WEBSOCKET_OPCODE_BINARY && (bits & 0xf) != MG_WEBSOCKET_OPCODE_CONTINUATION)
    {
        std::cerr << "Unknown opcode: failing." << std::endl;
        return false;
    }

    FILE *currFile = fopen(filename, "ab");
    fwrite(data, sizeof(char), data_len, currFile);
    fclose(currFile);

    return true;
}

void PrsonaServerWebSocketHandler::generate_response(
    struct mg_connection *conn,
    char *filename)
{
    const struct mg_request_info *info = mg_get_request_info(conn);

    switch (info->query_string[0])
    {
        case PRSONA_VERIFY_REPUTATION_PROOF:
            verify_reputation_proof(conn, filename);
            break;

        default:
            break;
    }
}

virtual void PrsonaClientWebSocketHandler::handleClose(
    CivetServer *server,
    const struct mg_connection *conn)
{
    char *filename = (char *) mg_get_user_connection_data(conn);
    if (!filename)
        return;

    remove(filename);
    delete filename;
}

void PrsonaClientWebSocketHandler::verify_reputation_proof(
    struct mg_connection *conn, const char *filename) const
{
    ifstream file(filename);

    std::vector<Proof> pi;
    Twistpoint shortTermPublicKey;
    Scalar threshold;

    BinarySizeT sizeOfVector;
    file >> sizeOfVector;
    for (size_t i = 0; i < sizeOfVector.val(); i++)
    {
        Proof currProof;
        file >> currProof;

        pi.push_back(currProof);
    }
    file >> shortTermPublicKey;
    file >> threshold;

    std::vector<Proof> generatorProof;
    Twistpoint freshGenerator =
        get_generator(
            &generator,
            serverIPs,
            generatorProof,
            true);

    std::vector<Proof> encryptedScoreProof;
    EGCiphertext encryptedScore =
        get_server_committed_val<EGCiphertext>(
            &generator,
            serverIPs,
            encryptedScoreProof,
            shortTermPublicKey,
            USER_TALLY_URI,
            USER_TALLY_COMMIT_URI);

    bool flag =
        prsonaClient->verify_reputation_proof(
            pi,
            shortTermPublicKey,
            threshold,
            encryptedScoreProof,
            encryptedScore);

    string data = flag ? "\x01" : "\x00";

    mg_websocket_write(conn, MG_WEBSOCKET_OPCODE_BINARY, data.c_str(), 1);
    mg_websocket_write(conn, MG_WEBSOCKET_OPCODE_DATACOMPLETE, "", 0);
}
